/*
 * radio.c
 *
 *  Created on: Nov 5, 2011
 *      Author: tkl
 */
#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

#include <avr/io.h>
#include <avr/interrupt.h>

#include "cc1100.h"
#include "cc1100_rf_settings.h"
#include "radio.h"

#include <util/delay.h>

#define RADIO_MODE_IDLE 0x00
#define RADIO_MODE_RX   0x01
#define RADIO_MODE_TX   0x02

uint8_t rx_buffer[RX_BUFFER_SIZE];

static volatile uint8_t radio_mode = RADIO_MODE_IDLE;
static volatile bool    rx_active = false;
static volatile bool    tx_active = false;
//-----------------------------------------------------------------------------
static uint8_t spi_write(uint8_t addr, const uint8_t* buffer, uint8_t size);
static uint8_t spi_read(uint8_t addr, uint8_t* buffer, uint8_t size);

//-----------------------------------------------------------------------------
void setup_radio(void) {
    uint8_t i;
    // configure io
    DDRB = (DDRB |  (1 << DDB2)
    		     |  (1 << DDB1)
    		     |  (1 << DDB0))
    		     & ~(1 << DDB3);
    PORTB |= (1 << PB0);    // disable CS

    SPCR = (SPCR |  (1 << SPE)
    		     |  (1 << MSTR)
    		     |  (1 << SPR1)
    		     |  (1 << SPR0))
    		     & ~(1 << SPIE)
    		     & ~(1 << DORD)
    		     & ~(1 << CPOL)
    		     & ~(1 << CPHA);

    PORTB |= (1 << PB0);    // disable CS
    _delay_us(1);
    PORTB &= ~(1 << PB0); // enable CS
    _delay_us(1);
    PORTB |= (1 << PB0);    // disable CS
    _delay_us(41);
    PORTB &= ~(1 << PB0); // enable CS
    while(PORTB & (1 << PB3));  // wait for MISO
    _delay_us(50);
    spi_write(CC1100_SRES, NULL, 0);
    while(PORTB & (1 << PB3));  // wait for MISO
    _delay_us(50);
    PORTB |= (1 << PB0);    // disable CS

    spi_write(CC1100_SRES, NULL, 0);
    radio_mode = RADIO_MODE_IDLE;
    spi_write(CC1100_SIDLE, NULL, 0);

    const uint8_t rf_base[32][2] = {
    		{CC1100_FSCTRL1, SMARTRF_SETTING_FSCTRL1},
    		{CC1100_FSCTRL0, SMARTRF_SETTING_FSCTRL0},
    		{CC1100_FREQ2, SMARTRF_SETTING_FREQ2},
    		{CC1100_FREQ1, SMARTRF_SETTING_FREQ1},
    		{CC1100_FREQ0, SMARTRF_SETTING_FREQ0},
    		{CC1100_MDMCFG4, SMARTRF_SETTING_MDMCFG4},
    		{CC1100_MDMCFG3, SMARTRF_SETTING_MDMCFG3},
    		{CC1100_MDMCFG2, SMARTRF_SETTING_MDMCFG2},
    		{CC1100_MDMCFG1, SMARTRF_SETTING_MDMCFG1},
    		{CC1100_MDMCFG0, SMARTRF_SETTING_MDMCFG0},
    		{CC1100_CHANNR, SMARTRF_SETTING_CHANNR},
    		{CC1100_DEVIATN, SMARTRF_SETTING_DEVIATN},
    		{CC1100_FREND1, SMARTRF_SETTING_FREND1},
    		{CC1100_FREND0, SMARTRF_SETTING_FREND0},
    		{CC1100_MCSM0, SMARTRF_SETTING_MCSM0},
    		{CC1100_FOCCFG, SMARTRF_SETTING_FOCCFG},
    		{CC1100_BSCFG, SMARTRF_SETTING_BSCFG},
    		{CC1100_AGCCTRL2, SMARTRF_SETTING_AGCCTRL2},
    		{CC1100_AGCCTRL1, SMARTRF_SETTING_AGCCTRL1},
    		{CC1100_AGCCTRL0, SMARTRF_SETTING_AGCCTRL0},
    		{CC1100_FSCAL3, SMARTRF_SETTING_FSCAL3},
    		{CC1100_FSCAL2, SMARTRF_SETTING_FSCAL2},
    		{CC1100_FSCAL1, SMARTRF_SETTING_FSCAL1},
    		{CC1100_FSCAL0, SMARTRF_SETTING_FSCAL0},
    		{CC1100_FSTEST, SMARTRF_SETTING_FSTEST},
    		{CC1100_TEST2, SMARTRF_SETTING_TEST2},
    		{CC1100_TEST1, SMARTRF_SETTING_TEST1},
    		{CC1100_TEST0, SMARTRF_SETTING_TEST0},
    		{CC1100_IOCFG2, SMARTRF_SETTING_IOCFG2},
    		{CC1100_PKTCTRL1, SMARTRF_SETTING_PKTCTRL1},
    		{CC1100_PKTCTRL0, SMARTRF_SETTING_PKTCTRL0},
    		{CC1100_PKTLEN, SMARTRF_SETTING_PKTLEN}
    };

    for(i = 0; i < 32; i++) {
    	spi_write(rf_base[i][0], &rf_base[i][1], 1);
    }

    // init gdo0
    uint8_t tmp = 0x06;
    spi_write(CC1100_IOCFG0, &tmp, 1);
    DDRD &= ~(1 << DDD1);   // GDO0 as input
    PORTD |= (1 << PD1);    // enable pull up
    EICRA = (1 << ISC01);  // falling edge
    EIMSK = (1 << INT0);

    radio_mode = RADIO_MODE_RX;
    spi_write(CC1100_SRX, NULL, 0);
}

//-----------------------------------------------------------------------------
uint8_t rx_packet(uint8_t *buffer, uint8_t len, uint32_t timeout) {
    uint8_t rx_bytes = 0;
    uint32_t count;

    for(count = 0; count < timeout; count++) {
        if(rx_active) {
            radio_mode = RADIO_MODE_IDLE;
            spi_write(CC1100_SIDLE, NULL, 0);

            spi_read((CC1100_RXBYTES | CC1100_READ_BURST), &rx_bytes, 1);
            if (rx_bytes > 0) {
                rx_bytes = spi_read(CC1100_RXFIFO | CC1100_READ_BURST, buffer, len);
                rx_bytes = buffer[0];
                spi_write(CC1100_SFRX, NULL, 0);
            }
            radio_mode = RADIO_MODE_RX;
            spi_write(CC1100_SRX, NULL, 0);

            rx_active = false;
            count = timeout;
        }
    }

    return rx_bytes;
}

//-----------------------------------------------------------------------------
void tx_packet(uint8_t *buffer, uint8_t len) {
    uint32_t count;
    if((buffer == NULL) || (len == 0)) {
        return;
    }
    uint8_t cPaTable = 0xC0;

    radio_mode = RADIO_MODE_IDLE;
    spi_write(CC1100_SIDLE, NULL, 0);

    spi_write(CC1100_PATABLE | CC1100_WRITE_BURST, &cPaTable, 1);
    spi_write(CC1100_SFTX, NULL, 0);

    spi_write(CC1100_TXFIFO | CC1100_WRITE_BURST, buffer, len);
    tx_active = true;

    EIMSK &= ~(1 << INT0);
    uint8_t tmp = 0x00;
    spi_write(CC1100_IOCFG0, &tmp, 1);

    radio_mode = RADIO_MODE_TX;
    spi_write(CC1100_STX, NULL, 0);

    EIMSK = (1 << INT0);
    tmp = 0x06;
    spi_write(CC1100_IOCFG0, &tmp, 1);
    // wait for transmission complete
//    while(tx_active);
    for(count = 0; count < 100000; count++) {
        if(!tx_active) {
            count = 100000;
        }
    }
    tx_active = false;

    radio_mode = RADIO_MODE_IDLE;
    spi_write(CC1100_SIDLE, NULL, 0);
    spi_write(CC1100_SFTX, NULL, 0);

    radio_mode = RADIO_MODE_RX;
    spi_write(CC1100_SRX, NULL, 0);
}

//-----------------------------------------------------------------------------
static uint8_t spi_write(uint8_t addr, const uint8_t* buffer, uint8_t size) {
    uint32_t wc = 0;
    uint8_t ret;
    uint8_t count;
    if((size > 0) && (buffer == NULL)) {
        return 0;
    }
    while(PORTB & (1 << PB3)) {  // wait for MISO
        wc++;
        if(wc > 100000) {
            return 0;
        }
    }
    PORTB &= ~(1 << PB0); // enable CS
    SPDR = addr;
    while(!(SPSR & (1 << SPIF))) {
        wc++;
        if(wc > 100000) {
            return 0;
        }
    }
    ret = SPDR;
    for(count = 0; count < size; count++) {
        SPDR = buffer[count];
        while(!(SPSR & (1 << SPIF))) {
            wc++;
            if(wc > 100000) {
                return 0;
            }
        }
    }
    PORTB |= (1 << PB0);    // disable CS
    return ret;
}

//-----------------------------------------------------------------------------
static uint8_t spi_read(uint8_t addr, uint8_t* buffer, uint8_t size) {
    uint32_t wc = 0;
    uint8_t ret;
    unsigned int i;
    PORTB &= ~(1 << PB0); // enable CS
//    while(PORTB & (1 << PB3));  // wait for MISO
    SPDR = addr;
    while(!(SPSR & (1 << SPIF))) {
        wc++;
        if(wc > 100000) {
            return 0;
        }
    }
    ret = SPDR;
    for(i = 0; i < size; i++) {
        SPDR = 0;
        while(!(SPSR & (1 << SPIF))) {
            wc++;
            if(wc > 100000) {
                return 0;
            }
        }
        buffer[i] = SPDR;
    }
    PORTB |= (1 << PB0);    // disable CS
    return ret;
}

//-----------------------------------------------------------------------------
// Interrupt receive routine for the wakeup rx task
//-----------------------------------------------------------------------------
ISR(INT0_vect) {
    switch (radio_mode) {
    case RADIO_MODE_RX:
        rx_active = true;
        break;

    case RADIO_MODE_TX:
        tx_active = false;
        break;
    }
}
